@@ -51,6 +51,12 @@ poppler_action_get_type (void)
   return our_type;
 }
 
+/**
+ * poppler_action_free:
+ * @action: a #PopplerAction
+ * 
+ * Frees @action
+ **/
 void
 poppler_action_free (PopplerAction *action)
 {
@@ -71,6 +77,14 @@ poppler_action_free (PopplerAction *action)
 	g_free (action);
 }
 
+/**
+ * poppler_action_copy:
+ * @action: a #PopplerAction
+ * 
+ * Copies @action, creating an identical #PopplerAction.
+ * 
+ * Return value: a new action identical to @action
+ **/
 PopplerAction *
 poppler_action_copy (PopplerAction *action)
 {
@@ -209,9 +223,14 @@ build_goto_remote (PopplerAction *action,
 
 static void
 build_launch (PopplerAction *action,
-	      LinkAction    *link)
+	      LinkLaunch    *link)
 {
-	/* FIXME: Write */
+	if (link->getFileName()) {
+		action->launch.file_name = link->getFileName()->getCString ();
+	}
+	if (link->getParams()) {
+		action->launch.file_name = link->getParams()->getCString ();
+	}
 }
 
 static void
@@ -248,6 +267,9 @@ _poppler_action_new (PopplerDocument *document,
 
 	action = g_new0 (PopplerAction, 1);
 
+	if (title)
+		action->any.title = g_strdup (title);
+
 	if (link == NULL) {
 		action->type = POPPLER_ACTION_UNKNOWN;
 		return action;
@@ -264,7 +286,7 @@ _poppler_action_new (PopplerDocument *document,
 		break;
 	case actionLaunch:
 		action->type = POPPLER_ACTION_LAUNCH;
-		build_launch (action, link);
+		build_launch (action, dynamic_cast <LinkLaunch *> (link));
 		break;
 	case actionURI:
 		action->type = POPPLER_ACTION_URI;
@@ -284,7 +306,5 @@ _poppler_action_new (PopplerDocument *document,
 		break;
 	}
 
-	if (title)
-		action->any.title = g_strdup (title);
 	return action;
 }
