@@ -73,6 +73,14 @@ poppler_page_finalize (GObject *object)
   /* page->page is owned by the document */
 }
 
+/**
+ * poppler_page_get_size:
+ * @page: A #PopplerPage
+ * @width: return location for the width of @page
+ * @height: return location for the height of @page
+ * 
+ * Gets the size of @page at the current scale and rotation.
+ **/
 void
 poppler_page_get_size (PopplerPage *page,
 		       double      *width,
@@ -85,11 +93,11 @@ poppler_page_get_size (PopplerPage *page,
 
   rotate = page->page->getRotate ();
   if (rotate == 90 || rotate == 270) {
-    page_height = page->page->getWidth ();
-    page_width = page->page->getHeight ();
+    page_height = page->page->getCropWidth ();
+    page_width = page->page->getCropHeight ();
   } else {
-    page_width = page->page->getWidth ();
-    page_height = page->page->getHeight ();
+    page_width = page->page->getCropWidth ();
+    page_height = page->page->getCropHeight ();
   }
 
   if (width != NULL)
@@ -98,6 +106,14 @@ poppler_page_get_size (PopplerPage *page,
     *height = page_height;
 }
 
+/**
+ * poppler_page_get_index:
+ * @page: a #PopplerPage
+ * 
+ * Returns the index of @page
+ * 
+ * Return value: index value of @page
+ **/
 int
 poppler_page_get_index (PopplerPage *page)
 {
@@ -129,11 +145,11 @@ poppler_page_prepare_output_dev (PopplerPage *page,
   rotate = (rotation + page->page->getRotate()) % 360;
 
   if (rotate == 90 || rotate == 270) {
-    cairo_width = MAX ((int)(page->page->getHeight() * scale + 0.5), 1);
-    cairo_height = MAX ((int)(page->page->getWidth() * scale + 0.5), 1);
+    cairo_width = MAX ((int)(page->page->getCropHeight() * scale + 0.5), 1);
+    cairo_height = MAX ((int)(page->page->getCropWidth() * scale + 0.5), 1);
   } else {
-    cairo_width = MAX ((int)(page->page->getWidth() * scale + 0.5), 1);
-    cairo_height = MAX ((int)(page->page->getHeight() * scale + 0.5), 1);
+    cairo_width = MAX ((int)(page->page->getCropWidth() * scale + 0.5), 1);
+    cairo_height = MAX ((int)(page->page->getCropHeight() * scale + 0.5), 1);
   }
 
   output_dev = page->document->output_dev;
@@ -246,23 +262,22 @@ poppler_page_copy_to_pixbuf(PopplerPage *page,
   if (splash_height > gdk_pixbuf_get_height (pixbuf))
     splash_height = gdk_pixbuf_get_height (pixbuf);
 
+  SplashColorPtr pixel = new Guchar[4];
   for (y = 0; y < splash_height; y++)
+  {
+    dst = pixbuf_data + y * pixbuf_rowstride;
+    for (x = 0; x < splash_width; x++)
     {
-      SplashRGB8 *src;
-
-      src = (SplashRGB8 *) (color_ptr.rgb8p + y * splash_rowstride);
-      dst = pixbuf_data + y * pixbuf_rowstride;
-      for (x = 0; x < splash_width; x++) 
-	{
-	  dst[0] = splashRGB8R(*src);
-	  dst[1] = splashRGB8G(*src); 
-	  dst[2] = splashRGB8B(*src);
-	  if (pixbuf_n_channels == 4)
-	    dst[3] = 0xff;
-	  dst += pixbuf_n_channels;
-	  src++;
-	}
+      output_dev->getBitmap()->getPixel(x, y, pixel);
+      dst[0] = pixel[0];
+      dst[1] = pixel[1];
+      dst[2] = pixel[2];
+      if (pixbuf_n_channels == 4)
+        dst[3] = 0xff;
+      dst += pixbuf_n_channels;
     }
+  }
+  delete pixel;
 }
 
 #endif
@@ -301,6 +316,7 @@ poppler_page_render_to_pixbuf (PopplerPage *page,
   page->page->displaySlice(page->document->output_dev,
 			   72.0 * scale, 72.0 * scale,
 			   rotation,
+			   gFalse, /* useMediaBox */
 			   gTrue, /* Crop */
 			   src_x, src_y,
 			   src_width, src_height,
@@ -318,6 +334,7 @@ poppler_page_get_text_output_dev (PopplerPage *page)
 
     page->gfx = page->page->createGfx(page->text_dev,
 				      72.0, 72.0, 0,
+				      gFalse, /* useMediaBox */
 				      gTrue, /* Crop */
 				      -1, -1, -1, -1,
 				      NULL, /* links */
@@ -332,6 +349,18 @@ poppler_page_get_text_output_dev (PopplerPage *page)
   return page->text_dev;
 }
 
+/**
+ * poppler_page_get_selection_region:
+ * @page: a #PopplerPage
+ * @scale: scale specified as pixels per point
+ * @selection: start and end point of selection as a rectangle
+ * 
+ * Returns a region containing the area that would be rendered by
+ * poppler_page_render_selection().  The returned region must be freed with
+ * gdk_region_destroy().
+ * 
+ * Return value: a newly allocated #GdkRegion
+ **/
 GdkRegion *
 poppler_page_get_selection_region (PopplerPage      *page,
 				   gdouble           scale,
@@ -389,7 +418,7 @@ poppler_page_set_selection_alpha (PopplerPage      *page,
 				  PopplerRectangle *selection)
 {
   GdkRegion *region;
-  gint n_rectangles, i, x, y;
+  gint n_rectangles, i, x, y, width, height;
   GdkRectangle *rectangles;
   int pixbuf_rowstride, pixbuf_n_channels;
   guchar *pixbuf_data, *dst;
@@ -397,10 +426,20 @@ poppler_page_set_selection_alpha (PopplerPage      *page,
   pixbuf_data = gdk_pixbuf_get_pixels (pixbuf);
   pixbuf_rowstride = gdk_pixbuf_get_rowstride (pixbuf);
   pixbuf_n_channels = gdk_pixbuf_get_n_channels (pixbuf);
+  width = gdk_pixbuf_get_width (pixbuf);
+  height = gdk_pixbuf_get_height (pixbuf);
 
   if (pixbuf_n_channels != 4)
     return;
 
+  for (y = 0; y < height; y++) {
+      dst = pixbuf_data + y * pixbuf_rowstride;
+      for (x = 0; x < width; x++) {
+	  dst[3] = 0x00;
+	  dst += pixbuf_n_channels;
+      }
+  }
+
   region = poppler_page_get_selection_region (page, scale, selection);
 
   gdk_region_get_rectangles (region, &rectangles, &n_rectangles);
@@ -459,14 +498,14 @@ poppler_page_render_selection (PopplerPage      *page,
 			     selection->x2, selection->y2);
 
   GfxColor gfx_background_color = { 
-    background_color->red / 65535.0,
-    background_color->green / 65535.0,
-    background_color->blue / 65535.0
+    background_color->red,
+    background_color->green,
+    background_color->blue
   };
   GfxColor gfx_glyph_color = {
-    glyph_color->red / 65535.0,
-    glyph_color->green / 65535.0,
-    glyph_color->blue / 65535.0
+    glyph_color->red,
+    glyph_color->green,
+    glyph_color->blue
   };
 
   text_dev = poppler_page_get_text_output_dev (page);
@@ -642,15 +681,17 @@ poppler_page_find_text (PopplerPage *page,
   doc = page->document->doc;
 
   poppler_page_get_size (page, NULL, &height);
-  page->page->display (output_dev, 72, 72, 0,
+  page->page->display (output_dev, 72, 72, 0, gFalse,
 		       gTrue, NULL, doc->getCatalog());
   
   matches = NULL;
   xMin = 0;
   yMin = 0;
+#warning you probably want to add caseSensitive and backwards as parameters
   while (output_dev->findText (ucs4, ucs4_len,
 			       gFalse, gTrue, // startAtTop, stopAtBottom
 			       gTrue, gFalse, // startAtLast, stopAtLast
+			       gFalse, gFalse, // caseSensitive, backwards
 			       &xMin, &yMin, &xMax, &yMax))
     {
       match = g_new (PopplerRectangle, 1);
@@ -686,14 +727,14 @@ poppler_page_render_to_ps (PopplerPage   *page,
     ps_file->out = new PSOutputDev (ps_file->filename,
                                     ps_file->document->doc->getXRef(),
                                     ps_file->document->doc->getCatalog(),
-                                    ps_file->first_page + 1, ps_file->last_page,
+                                    ps_file->first_page, ps_file->last_page,
                                     psModePS, (int)ps_file->paper_width,
                                     (int)ps_file->paper_height, ps_file->duplex,
                                     0, 0, 0, 0, gFalse);
 
 
   ps_file->document->doc->displayPage (ps_file->out, page->index + 1, 72.0, 72.0,
-				       0, gTrue, gFalse);
+				       0, gFalse, gTrue, gFalse);
 }
 
 static void
@@ -780,10 +821,10 @@ poppler_page_get_link_mapping (PopplerPage *page)
 		link->getRect (&(mapping->area.x1), &(mapping->area.y1),
 			       &(mapping->area.x2), &(mapping->area.y2));
 
-		mapping->area.x1 -= page->page->getBox()->x1;
-		mapping->area.x2 -= page->page->getBox()->x1;
-		mapping->area.y1 -= page->page->getBox()->y1;
-		mapping->area.y2 -= page->page->getBox()->y1;
+		mapping->area.x1 -= page->page->getCropBox()->x1;
+		mapping->area.x2 -= page->page->getCropBox()->x1;
+		mapping->area.y1 -= page->page->getCropBox()->y1;
+		mapping->area.y2 -= page->page->getCropBox()->y1;
 
 		map_list = g_list_prepend (map_list, mapping);
 	}
