@@ -33,9 +33,7 @@
 
 //------------------------------------------------------------------------
 
-#define soutRound(x) ((int)(x + 0.5))
-
-//#define LOG_CAIRO
+// #define LOG_CAIRO
 
 #ifdef LOG_CAIRO
 #define LOG(x) (x)
@@ -55,6 +53,11 @@ CairoOutputDev::CairoOutputDev() {
   fontEngine = NULL;
   glyphs = NULL;
   surface = NULL;
+  fill_pattern = NULL;
+  stroke_pattern = NULL;
+  stroke_opacity = 1.0;
+  fill_opacity = 1.0;
+  textClipPath = NULL;
 }
 
 CairoOutputDev::~CairoOutputDev() {
@@ -63,6 +66,8 @@ CairoOutputDev::~CairoOutputDev() {
   }
   FT_Done_FreeType(ft_lib);
   cairo_surface_destroy (surface);
+  cairo_pattern_destroy (stroke_pattern);
+  cairo_pattern_destroy (fill_pattern);
 }
 
 void CairoOutputDev::setSurface(cairo_surface_t *surface)
@@ -194,21 +199,51 @@ void CairoOutputDev::updateLineWidth(GfxState *state) {
 
 void CairoOutputDev::updateFillColor(GfxState *state) {
   state->getFillRGB(&fill_color);
-  LOG(printf ("fill color: %f %f %f\n", fill_color.r, fill_color.g, fill_color.b));
+
+  cairo_pattern_destroy(fill_pattern);
+  fill_pattern = cairo_pattern_create_rgba(fill_color.r / 65535.0,
+					   fill_color.g / 65535.0,
+					   fill_color.b / 65535.0,
+					   fill_opacity);
+
+  LOG(printf ("fill color: %d %d %d\n",
+	      fill_color.r, fill_color.g, fill_color.b));
 }
 
 void CairoOutputDev::updateStrokeColor(GfxState *state) {
   state->getStrokeRGB(&stroke_color);
-  LOG(printf ("stroke color: %f %f %f\n", stroke_color.r, stroke_color.g, stroke_color.b));
+
+  cairo_pattern_destroy(stroke_pattern);
+  stroke_pattern = cairo_pattern_create_rgba(stroke_color.r / 65535.0,
+					     stroke_color.g / 65535.0,
+					     stroke_color.b / 65535.0,
+					     stroke_opacity);
+  
+  LOG(printf ("stroke color: %d %d %d\n",
+	      stroke_color.r, stroke_color.g, stroke_color.b));
 }
 
 void CairoOutputDev::updateFillOpacity(GfxState *state) {
   fill_opacity = state->getFillOpacity();
+
+  cairo_pattern_destroy(fill_pattern);
+  fill_pattern = cairo_pattern_create_rgba(fill_color.r / 65535.0,
+					   fill_color.g / 65535.0,
+					   fill_color.b / 65535.0,
+					   fill_opacity);
+
   LOG(printf ("fill opacity: %f\n", fill_opacity));
 }
 
 void CairoOutputDev::updateStrokeOpacity(GfxState *state) {
   stroke_opacity = state->getStrokeOpacity();
+
+  cairo_pattern_destroy(stroke_pattern);
+  stroke_pattern = cairo_pattern_create_rgba(stroke_color.r / 65535.0,
+					     stroke_color.g / 65535.0,
+					     stroke_color.b / 65535.0,
+					     stroke_opacity);
+  
   LOG(printf ("stroke opacity: %f\n", stroke_opacity));
 }
 
@@ -219,19 +254,18 @@ void CairoOutputDev::updateFont(GfxState *state) {
   cairo_matrix_t matrix;
 
   LOG(printf ("updateFont() font=%s\n", state->getFont()->getName()->getCString()));
-  
+
   needFontUpdate = gFalse;
 
+  if (state->getFont()->getType() == fontType3)
+    return;
+
   currentFont = fontEngine->getFont (state->getFont(), xref);
 
   state->getFontTransMat(&m11, &m12, &m21, &m22);
   m11 *= state->getHorizScaling();
   m12 *= state->getHorizScaling();
 
-  w = currentFont->getSubstitutionCorrection(state->getFont());
-  m12 *= w;
-  m22 *= w;
-
   LOG(printf ("font matrix: %f %f %f %f\n", m11, m12, m21, m22));
   
   font_face = currentFont->getFontFace();
@@ -286,9 +320,7 @@ void CairoOutputDev::doPath(GfxState *state, GfxPath *path) {
 
 void CairoOutputDev::stroke(GfxState *state) {
   doPath (state, state->getPath());
-  cairo_set_source_rgba (cairo,
-			 stroke_color.r, stroke_color.g, stroke_color.b,
-			 stroke_opacity);
+  cairo_set_source (cairo, stroke_pattern);
   LOG(printf ("stroke\n"));
   cairo_stroke (cairo);
 }
@@ -296,9 +328,7 @@ void CairoOutputDev::stroke(GfxState *state) {
 void CairoOutputDev::fill(GfxState *state) {
   doPath (state, state->getPath());
   cairo_set_fill_rule (cairo, CAIRO_FILL_RULE_WINDING);
-  cairo_set_source_rgba (cairo,
-			 fill_color.r, fill_color.g, fill_color.b,
-			 fill_opacity);
+  cairo_set_source (cairo, fill_pattern);
   LOG(printf ("fill\n"));
   cairo_fill (cairo);
 }
@@ -306,8 +336,7 @@ void CairoOutputDev::fill(GfxState *state) {
 void CairoOutputDev::eoFill(GfxState *state) {
   doPath (state, state->getPath());
   cairo_set_fill_rule (cairo, CAIRO_FILL_RULE_EVEN_ODD);
-  cairo_set_source_rgb (cairo,
-		       fill_color.r, fill_color.g, fill_color.b);
+  cairo_set_source (cairo, fill_pattern);
   LOG(printf ("fill-eo\n"));
   cairo_fill (cairo);
 }
@@ -340,7 +369,7 @@ void CairoOutputDev::beginString(GfxState *state, GooString *s)
 void CairoOutputDev::drawChar(GfxState *state, double x, double y,
 			      double dx, double dy,
 			      double originX, double originY,
-			      CharCode code, Unicode *u, int uLen)
+			      CharCode code, int nBytes, Unicode *u, int uLen)
 {
   double tx, ty;
 
@@ -358,13 +387,10 @@ void CairoOutputDev::endString(GfxState *state)
   if (!currentFont)
     return;
    
-  // check for invisible text -- this is used by Acrobat Capture
+  // ignore empty strings and invisible text -- this is used by
+  // Acrobat Capture
   render = state->getRender();
-  if (render == 3)
-    return;
-
-  // ignore empty strings
-  if (glyphCount == 0) {
+  if (render == 3 || glyphCount == 0) {
     gfree(glyphs);
     glyphs = NULL;
     return;
@@ -372,28 +398,36 @@ void CairoOutputDev::endString(GfxState *state)
   
   if (!(render & 1)) {
     LOG (printf ("fill string\n"));
-    cairo_set_source_rgb (cairo,
-			 fill_color.r, fill_color.g, fill_color.b);
+    cairo_set_source (cairo, fill_pattern);
     cairo_show_glyphs (cairo, glyphs, glyphCount);
   }
   
   // stroke
   if ((render & 3) == 1 || (render & 3) == 2) {
     LOG (printf ("stroke string\n"));
-    cairo_set_source_rgb (cairo,
-			 stroke_color.r, stroke_color.g, stroke_color.b);
+    cairo_set_source (cairo, stroke_pattern);
     cairo_glyph_path (cairo, glyphs, glyphCount);
     cairo_stroke (cairo);
   }
 
   // clip
   if (render & 4) {
-    // FIXME: This is quite right yet, we need to accumulate all
-    // glyphs within one text object before we clip.  Right now this
-    // just add this one string.
     LOG (printf ("clip string\n"));
+    // append the glyph path to textClipPath.
+
+    // set textClipPath as the currentPath
+    if (textClipPath) {
+      cairo_append_path (cairo, textClipPath);
+      cairo_path_destroy (textClipPath);
+    }
+    
+    // append the glyph path
     cairo_glyph_path (cairo, glyphs, glyphCount);
-    cairo_clip (cairo);
+   
+    // move the path back into textClipPath 
+    // and clear the current path
+    textClipPath = cairo_copy_path (cairo);
+    cairo_new_path (cairo);
   }
   
   gfree (glyphs);
@@ -417,6 +451,14 @@ void CairoOutputDev::type3D1(GfxState *state, double wx, double wy,
 }
 
 void CairoOutputDev::endTextObject(GfxState *state) {
+  if (textClipPath) {
+    // clip the accumulated text path
+    cairo_append_path (cairo, textClipPath);
+    cairo_clip (cairo);
+    cairo_path_destroy (textClipPath);
+    textClipPath = NULL;
+  }
+
 }
 
 
@@ -482,7 +524,7 @@ void CairoOutputDev::drawImageMask(GfxState *state, Object *ref, Stream *str,
 
   cairo_pattern_set_filter (pattern, CAIRO_FILTER_BEST);
   /* FIXME: Doesn't the image mask support any colorspace? */
-  cairo_set_source_rgb (cairo, fill_color.r, fill_color.g, fill_color.b);
+  cairo_set_source (cairo, fill_pattern);
   cairo_mask (cairo, pattern);
 
   cairo_pattern_destroy (pattern);
