@@ -102,7 +102,7 @@ public:
   void drawChar(GfxState *state, double x, double y,
 		double dx, double dy,
 		double originX, double originY,
-		CharCode code, Unicode *u, int uLen);
+		CharCode code, int nBytes, Unicode *u, int uLen);
 
   virtual GBool beginType3Char(GfxState *state, double x, double y,
 			       double dx, double dy,
@@ -135,8 +135,8 @@ public:
 protected:
   void doPath(GfxState *state, GfxPath *path);
   
-  GfxRGB fill_color;
-  GfxRGB stroke_color;
+  GfxRGB fill_color, stroke_color;
+  cairo_pattern_t *fill_pattern, *stroke_pattern;
   double fill_opacity;
   double stroke_opacity;
   CairoFont *currentFont;
@@ -150,6 +150,7 @@ protected:
   cairo_surface_t *surface;
   cairo_glyph_t *glyphs;
   int glyphCount;
+  cairo_path_t *textClipPath;
 };
 
 #endif
