@@ -26,6 +26,7 @@ class GfxSeparationColorSpace;
 class PDFRectangle;
 struct PSFont16Enc;
 class PSOutCustomColor;
+class Function;
 
 //------------------------------------------------------------------------
 // PSOutputDev
@@ -83,6 +84,17 @@ public:
   // Does this device use drawChar() or drawString()?
   virtual GBool useDrawChar() { return gFalse; }
 
+  // Does this device use tilingPatternFill()?  If this returns false,
+  // tiling pattern fills will be reduced to a series of other drawing
+  // operations.
+  virtual GBool useTilingPatternFill() { return gTrue; }
+
+  // Does this device use functionShadedFill(), axialShadedFill(), and
+  // radialShadedFill()?  If this returns false, these shaded fills
+  // will be reduced to a series of other drawing operations.
+  virtual GBool useShadedFills()
+    { return level == psLevel2 || level == psLevel3; }
+
   // Does this device use beginType3Char/endType3Char?  Otherwise,
   // text in Type 3 fonts will be drawn with drawChar/drawString.
   virtual GBool interpretType3Chars() { return gFalse; }
@@ -91,7 +103,8 @@ public:
 
   // Write the document-level header.
   void writeHeader(int firstPage, int lastPage,
-		   PDFRectangle *mediaBox, PDFRectangle *cropBox);
+		   PDFRectangle *mediaBox, PDFRectangle *cropBox,
+		   int pageRotate);
 
   // Write the Xpdf procset.
   void writeXpdfProcset();
@@ -99,9 +112,6 @@ public:
   // Write the document-level setup.
   void writeDocSetup(Catalog *catalog, int firstPage, int lastPage, GBool duplexA);
 
-  // Write the setup for the current page.
-  void writePageSetup();
-
   // Write the trailer for the current page.
   void writePageTrailer();
 
@@ -129,8 +139,12 @@ public:
   virtual void updateLineCap(GfxState *state);
   virtual void updateMiterLimit(GfxState *state);
   virtual void updateLineWidth(GfxState *state);
+  virtual void updateFillColorSpace(GfxState *state);
+  virtual void updateStrokeColorSpace(GfxState *state);
   virtual void updateFillColor(GfxState *state);
   virtual void updateStrokeColor(GfxState *state);
+  virtual void updateFillOverprint(GfxState *state);
+  virtual void updateStrokeOverprint(GfxState *state);
 
   //----- update text state
   virtual void updateFont(GfxState *state);
@@ -147,6 +161,15 @@ public:
   virtual void stroke(GfxState *state);
   virtual void fill(GfxState *state);
   virtual void eoFill(GfxState *state);
+  virtual void tilingPatternFill(GfxState *state, Object *str,
+				 int paintType, Dict *resDict,
+				 double *mat, double *bbox,
+				 int x0, int y0, int x1, int y1,
+				 double xStep, double yStep);
+  virtual void functionShadedFill(GfxState *state,
+				  GfxFunctionShading *shading);
+  virtual void axialShadedFill(GfxState *state, GfxAxialShading *shading);
+  virtual void radialShadedFill(GfxState *state, GfxRadialShading *shading);
 
   //----- path clipping
   virtual void clip(GfxState *state);
@@ -163,6 +186,11 @@ public:
   virtual void drawImage(GfxState *state, Object *ref, Stream *str,
 			 int width, int height, GfxImageColorMap *colorMap,
 			 int *maskColors, GBool inlineImg);
+  virtual void drawMaskedImage(GfxState *state, Object *ref, Stream *str,
+			       int width, int height,
+			       GfxImageColorMap *colorMap,
+			       Stream *maskStr, int maskWidth, int maskHeight,
+			       GBool maskInvert);
 
 #if OPI_SUPPORT
   //----- OPI functions
@@ -211,7 +239,8 @@ private:
   void setupEmbeddedTrueTypeFont(GfxFont *font, Ref *id, GooString *psName);
   void setupExternalTrueTypeFont(GfxFont *font, GooString *psName);
   void setupEmbeddedCIDType0Font(GfxFont *font, Ref *id, GooString *psName);
-  void setupEmbeddedCIDTrueTypeFont(GfxFont *font, Ref *id, GooString *psName);
+  void setupEmbeddedCIDTrueTypeFont(GfxFont *font, Ref *id, GooString *psName,
+				    GBool needVerticalMetrics);
   void setupType3Font(GfxFont *font, GooString *psName, Dict *parentResDict);
   void setupImages(Dict *resDict);
   void setupImage(Ref id, Stream *str);
@@ -226,8 +255,11 @@ private:
 		    Stream *str, int width, int height, int len);
   void doImageL2(Object *ref, GfxImageColorMap *colorMap,
 		 GBool invert, GBool inlineImg,
-		 Stream *str, int width, int height, int len);
-  void dumpColorSpaceL2(GfxColorSpace *colorSpace);
+		 Stream *str, int width, int height, int len,
+		 int *maskColors, Stream *maskStr,
+		 int maskWidth, int maskHeight, GBool maskInvert);
+  void dumpColorSpaceL2(GfxColorSpace *colorSpace,
+			GBool genXform, GBool updateColors);
 #if OPI_SUPPORT
   void opiBegin20(GfxState *state, Dict *dict);
   void opiBegin13(GfxState *state, Dict *dict);
@@ -235,6 +267,7 @@ private:
 		    double *x1, double *y1);
   GBool getFileSpec(Object *fileSpec, Object *fileName);
 #endif
+  void cvtFunction(Function *func);
   void writePSChar(char c);
   void writePS(char *s);
   void writePSFmt(const char *fmt, ...) GCC_PRINTF_FORMAT(2, 3);
@@ -278,6 +311,8 @@ private:
   GooList *xobjStack;		// stack of XObject dicts currently being
 				//   processed
   int numSaves;			// current number of gsaves
+  int numTilingPatterns;	// current number of nested tiling patterns
+  int nextFunc;			// next unique number to use for a function
 
   double tx0, ty0;		// global translation
   double xScale0, yScale0;	// global scaling
@@ -287,6 +322,8 @@ private:
   double tx, ty;		// global translation for current page
   double xScale, yScale;	// global scaling for current page
   int rotate;			// rotation angle for current page
+  double epsX1, epsY1,		// EPS bounding box (unrotated)
+         epsX2, epsY2;
 
   GooString *embFontList;		// resource comments for embedded fonts
 
