@@ -1,5 +1,6 @@
 /* poppler-document.cc: qt interface to poppler
  * Copyright (C) 2005, Net Integration Technologies, Inc.
+ * Copyright (C) 2005, Brad Hards <bradh@frogmouth.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,14 +32,16 @@
 
 namespace Poppler {
 
-    Document *Document::load(const QString &filePath)
+  Document *Document::load(const QString &filePath, const QByteArray &ownerPassword,
+			   const QByteArray &userPassword)
     {
 	if (!globalParams) {
 	    globalParams = new GlobalParams("/etc/xpdfrc");
-	    globalParams->setupBaseFontsFc(NULL);
 	}
 
-	DocumentData *doc = new DocumentData(new GooString(QFile::encodeName(filePath)), NULL, NULL);
+	DocumentData *doc = new DocumentData(new GooString(QFile::encodeName(filePath)), 
+					     new GooString(ownerPassword.data()),
+					     new GooString(userPassword.data()));
 	Document *pdoc;
 	if (doc->doc.isOk() || doc->doc.getErrorCode() == errEncrypted) {
 	    pdoc = new Document(doc);
@@ -50,6 +53,7 @@ namespace Poppler {
 	    return pdoc;
 	}
 	else
+	    delete doc;
 	    return NULL;
 
     }
@@ -167,6 +171,7 @@ namespace Poppler {
 			      (Poppler::FontInfo::Type)((::FontInfo*)items->get(i))->getType()
 		));
 	}
+	deleteGooList(items, ::FontInfo);
 	return true;
     }
 
@@ -237,6 +242,28 @@ namespace Poppler {
 	return NULL;
     }
 
+    QStringList Document::infoKeys() const
+    {
+	QStringList keys;
+
+	Object info;
+	if ( m_doc->locked )
+	    return QStringList();
+
+	m_doc->doc.getDocInfo( &info );
+	if ( !info.isDict() )
+	    return QStringList();
+
+	Dict *infoDict = info.getDict();
+	// somehow iterate over keys in infoDict
+	for( int i=0; i < infoDict->getLength(); ++i ) {
+	    keys.append( QString::fromAscii(infoDict->getKey(i)) );
+	}
+
+	info.free();
+	return keys;
+    }
+
     /* borrowed from kpdf */
     QDateTime Document::date( const QString & type ) const
     {
@@ -255,43 +282,15 @@ namespace Poppler {
 	char *s;
 	int year, mon, day, hour, min, sec;
 	Dict *infoDict = info.getDict();
-	QString result;
+	QDateTime result;
 
 	if ( infoDict->lookup( type.toLatin1().data(), &obj )->isString() )
 	{
-	    s = obj.getString()->getCString();
-	    if ( s[0] == 'D' && s[1] == ':' )
-		s += 2;
-	    /* FIXME process time zone on systems that support it */  
-	    if ( sscanf( s, "%4d%2d%2d%2d%2d%2d", &year, &mon, &day, &hour, &min, &sec ) == 6 )
-	    {
-		/* Workaround for y2k bug in Distiller 3 stolen from gpdf, hoping that it won't
-		 *   * be used after y2.2k */
-		if ( year < 1930 && strlen (s) > 14) {
-		    int century, years_since_1900;
-		    if ( sscanf( s, "%2d%3d%2d%2d%2d%2d%2d",
-				 &century, &years_since_1900,
-				 &mon, &day, &hour, &min, &sec) == 7 )
-			year = century * 100 + years_since_1900;
-		    else {
-			obj.free();
-			info.free();
-			return QDateTime();
-		    }
-		}
-
-		QDate d( year, mon, day );  //CHECK: it was mon-1, Jan->0 (??)
-		QTime t( hour, min, sec );
-		if ( d.isValid() && t.isValid() ) {
-		    obj.free();
-		    info.free();
-		    return QDateTime( d, t );
-		}
-	    }
+	    result = Poppler::convertDate(obj.getString()->getCString());
 	}
 	obj.free();
 	info.free();
-	return QDateTime();
+	return result;
     }
 
     bool Document::isEncrypted() const
@@ -365,4 +364,59 @@ namespace Poppler {
 	return page(index);
     }
 
+    QDateTime convertDate( char *dateString )
+    {
+        int year;
+        int mon = 1;
+        int day = 1;
+        int hour = 0;
+        int min = 0;
+        int sec = 0;
+        char tz = 0x00;
+        int tzHours = 0;
+        int tzMins = 0;
+
+        if ( dateString[0] == 'D' && dateString[1] == ':' )
+            dateString += 2;
+        if ( sscanf( dateString,
+		     "%4d%2d%2d%2d%2d%2d%c%2d%*c%2d",
+		     &year, &mon, &day, &hour, &min, &sec,
+		     &tz, &tzHours, &tzMins ) > 0 ) {
+            /* Workaround for y2k bug in Distiller 3 stolen from gpdf, hoping that it won't
+             * be used after y2.2k */
+            if ( year < 1930 && strlen (dateString) > 14) {
+                int century, years_since_1900;
+                if ( sscanf( dateString, "%2d%3d%2d%2d%2d%2d%2d",
+                             &century, &years_since_1900,
+                             &mon, &day, &hour, &min, &sec) == 7 )
+                    year = century * 100 + years_since_1900;
+                else {
+                    return QDateTime();
+                }
+            }
+
+            QDate d( year, mon, day );
+            QTime t( hour, min, sec );
+            if ( d.isValid() && t.isValid() ) {
+                QDateTime dt( d, t, Qt::UTC );
+                if ( tz ) {
+                    // then we have some form of timezone
+                    if ( 'Z' == tz  ) {
+                        // We are already at UTC
+                    } else if ( '+' == tz ) {
+                        // local time is ahead of UTC
+                        dt = dt.addSecs(-1*((tzHours*60)+tzMins)*60);
+                    } else if ( '-' == tz ) {
+                        // local time is behind UTC
+                        dt = dt.addSecs(((tzHours*60)+tzMins)*60);
+                    } else {
+                        qWarning("unexpected tz val");
+                    }
+                }
+                return dt;
+            }
+        }
+        return QDateTime();
+    }
+
 }
